# Docker images

Docker images can be used with the docker command.
To make their usage a bit easier on Unix-like operating systems, a standalone script is also available.

The first section describes how to use this script, while the second describes low-level usage with docker.

A third section is dedicated to Singularity/Apptainer.

## The *larvatagger.sh* script

### Getting the script

The *larvatagger.sh* script is available in the *scripts* directory of the project.

If you do not have a copy of the project, you can simply [download the script](https://gitlab.pasteur.fr/nyx/larvatagger.jl/-/raw/main/scripts/larvatagger.sh?inline=false) and set its permissions so that it can be executed.

Note that Docker still has to be installed.

To check the script works properly, try:
```
./larvatagger.sh --version
```

### Launching the GUI

To open a track data file, try:
```
./larvatagger.sh open path/to/datafile
```
This turns the commandline into a Julia interpreter to serve the app as a webserver.

As instructed, open `localhost:9284` in your webbrowser to access the app.
As similarly instructed, you can stop the server in the commandline pressing Ctrl+D.
The server can also be exited typing `exit()` and pressing Enter.

A default tagger will be available in the app.

### Automatic tagging

Behavior labels files (*.json* files with extension *.label) can be generated using the CLI with:
```
./larvatagger.sh predict path/to/data/file --skip-make-dataset
```
Similarly, all the track data files in a directory tree can be processed with a single command:
```
./larvatagger.sh predict path/to/data/repository --skip-make-dataset
```

Training a new tagger is almost as easy:
```
./larvatagger.sh train path/to/data/repo my-tagger
```
However, for the trained models to be persistant, the `train` command must store them outside the container.
Per default, `larvatagger.sh train` creates a *models* directory in the current directory and stores the model instances in there.

As a consequence, the `predict` command can use external model instances passing paths instead of instance names.
To distinguish between paths and names, `larvatagger.sh` checks whether the corresponding directory exists. If a directory exists at the path pointed to by `--model-instance` (on the host), then it is mounted into the container as an external trained model:
```
./larvatagger.sh predict path/to/data/file --model-instance models/my-tagger --skip-make-dataset
```

To enable GPU acceleration, pass the `--gpus all` argument.
Note however the default MaggotUBA backend does not support GPU acceleration at the moment.

To check whether GPUs are detected, run `docker run --rm --gpus all --entrypoint checkgpu flaur/larvatagger`. This command should print a list of your GPUs, otherwise prints *cuda unavailable*.

### Building an image

The *larvatagger.sh* script features a `build` command to build a local Docker image:
```
scripts/larvatagger.sh build
```

Optionally, you can also get the default backend (currently *MaggotUBA/20221005-1*) with:
```
scripts/larvatagger.sh build --with-default-backend
```

If you specifically want the *20221005-1* tagger, and not rely on default, do:
```
scripts/larvatagger.sh build --with-backend=MaggotUBA/20221005-1
```

The `build` command requires a local copy of the LarvaTagger.jl project. In addition, the larvatagger.sh script must be executed from the project root directory, hence the path `scripts/larvatagger.sh`.

## Using the docker command

### Getting available images

Images are published on [Docker Hub](https://hub.docker.com/r/flaur/larvatagger/tags).

They ship with `LarvaTagger.jl` and, optionally, with a MaggotUBA-based toy backend for automatic tagging.

Try:
```
docker pull flaur/larvatagger
```

Beware that images that ship with backends are relatively large files (>5GB on disk).
If you are not interested in automatic tagging, use the `flaur/larvatagger:0.7.1-standalone` image instead.

### Upgrading

The `docker pull flaur/larvatagger` command can be used to upgrade the Docker image as well.

It should be safe to upgrade but beware that, for example, the default tagger may change.

The last release known to include a specific tagger will stay available on dockerhub.

### Launching the GUI

A number of options must be passed to the `docker` command for the container to open a data file:

```
docker -iv $(pwd):/data -p 9284:9284 flaur/larvatagger open /data/path/to/your/file
```
with `path/to/your/file` a relative path to your file.

In PowerShell on Windows, use `${PWD}` instead of `$(pwd)`.

In the command above, `/data` represents the local directory, as made available within the container.
This is required, since no files outside the container can be accessed.

The port LarvaTagger.jl listens to must also be exported with `-p 9284:9284` so that it can be accessed from outside the container.

### `train` and `predict`

Just like the *larvatagger.jl* script, the docker image admits more commands/switches, including `import`, `train` and `predict`, or options such as `--help`.
For these other commands, neither `-i` nor `-p 9284:9284` are necessary.

See the main [README](https://gitlab.pasteur.fr/nyx/larvatagger.jl#automatic-tagging) for usage information.

Additionally, for the `train` switch, an external *models* directory must be explicitly mounted */app/MaggotUBA/models*.
Ownership of the generated files will have to be fixed (*root* per default on Unix-like systems).

### Windows PowerShell example

Using a few [sample files](https://gitlab.pasteur.fr/nyx/artefacts/-/tree/master/PlanarLarvae):
```
PS D:\> ls | Select Name

Name
----
trx.mat
trx_small.mat


PS D:\> mkdir models
```
We train a *test-tagger* model instance for the *MaggotUBA* backend shipped with image *flaur/larvatagger:latest*:
```
PS D:\> docker run -v ${PWD}:/data --mount type=bind,src=D:/models,dst=/app/MaggotUBA/models flaur/larvatagger train /app/MaggotUBA /data/trx.mat test-tagger --labels run_large,cast_large,back_large,stop_large --layers 2
┌ Info: Pushing file to backend
│   backend = "MaggotUBA"
│   instance = "test-tagger"
└   file = "/data/trx.mat"
WARNING: could not import HDF5.exists into MAT
INFO:main: running make_dataset.py
[ Info: Counting behavior labels in run: test-tagger
┌ Info: Sample sizes (observed, selected):
│   stop_large = (392, 392)
│   cast_large = (30339, 784)
│   run_large = (68764, 784)
└   back_large = (1818, 392)
[ Info: Sampling series of spines in run: test-tagger
INFO:main: running train_model.py
DEBUG:train_model: trailing arguments: {'layers': '2'}
PS D:\> ls models/test-tagger | Select Name

Name
----
autoencoder_config.json
best_validated_encoder.pt
clf_config.json
retrained_encoder.pt
trained_classifier.pt

```
This results in a new directory in *models*.

We now tag a second data file using the *test-tagger* instance:
```
PS D:\> docker run -v ${PWD}:/data --mount type=bind,src=D:/models,dst=/app/MaggotUBA/models flaur/larvatagger predict /app/MaggotUBA test-tagger /data/trx_small.mat --skip-make-dataset
┌ Info: Pushing file to backend
│   backend = "MaggotUBA"
│   instance = "test-tagger"
└   file = "/data/trx_small.mat"
WARNING: could not import HDF5.exists into MAT
INFO:main: running predict_model.py
PS D:\> ls | Select Name

Name
----
models
predicted.label
trx.mat
trx_small.mat

```
A *predicted.label* file was generated.
This file can be loaded in the LarvaTagger.jl viewer like any *table.csv*, *.mat* or *.outline* file.

## Singularity/Apptainer

Docker is usually not available on shared HPC resources.
[Singularity](https://sylabs.io/) or [Apptainer](http://apptainer.org/) are used instead, and the Docker image for LarvaTagger.jl can be run using these technologies.

An issue with the `predict` and `train` switches makes the backend unable to use the internal */app/MaggotUBA/data* directory to organise the raw, interim and processed data files.
A workaround consists in using an external *data* directory and bind mount it as */app/MaggotUBA/data*.
For example:

```
apptainer run --bind data:/app/MaggotUBA/data docker://flaur/larvatagger predict /app/MaggotUBA 20221005-1 path/to/datafile --skip-make-dataset
```
