module Taggers

import PlanarLarvae.Formats

export Tagger, isbackend, resetmodel, resetdata, train, predict

struct Tagger
    backend_dir::String
    model_instance::String
end

isbackend(path) = isdir(joinpath(path, "models")) &&
                  isfile(joinpath(path, "pyproject.toml")) &&
                  isfile(joinpath(path, "poetry.lock"))
isbackend(tagger::Tagger) = isbackend(tagger.backend_dir)

modeldir(tagger::Tagger) = joinpath(tagger.backend_dir, "models", tagger.model_instance)

datadir(tagger::Tagger, stage::String) = joinpath(tagger.backend_dir, "data", stage, tagger.model_instance)

function reset(tagger::Tagger)
    resetmodel(tagger)
    resetdata(tagger)
end

function reset(dir::String)
    try
        rm(dir; recursive=true)
    catch
    end
    mkpath(dir)
    check_permissions(dir)
end

resetmodel(tagger::Tagger) = reset(modeldir(tagger))

function resetdata(tagger::Tagger)
    for dir in ("raw", "interim", "processed")
        resetdata(tagger, dir)
    end
end

resetdata(tagger::Tagger, dir::String) = reset(datadir(tagger, dir))

"""
    push(tagger, filepath)
    push(tagger, dirpath)
    push(tagger, filelistfile)

Copy a data file into the tagger's raw data directory,
or link a data directory as the tagger's raw data directory,
or read relative filepaths from a .txt file and copy all listed files
into the tagger's raw data directory, preserving the directory structure
in the relative paths.
"""
function push(tagger::Tagger, inputdata::String)
    destination = nothing
    raw_data_dir = datadir(tagger, "raw")
    if isdir(inputdata)
        inputdata = realpath(inputdata) # strip the end slashes
        rm(raw_data_dir; force=true, recursive=true)
        destination = raw_data_dir
        symlink(inputdata, destination)
    elseif endswith(inputdata, ".txt")
        files_by_dir = Dict{String, Vector{String}}()
        for file in readlines(inputdata)
            parent = dirname(file)
            push!(get!(files_by_dir, parent, String[]), abspath(file))
        end
        for (dir, files) in pairs(files_by_dir)
            files = filter(files)
            for file in files
                @debug "Pushing file to backend" backend=basename(tagger.backend_dir) instance=tagger.model_instance file=file
                dest_dir = normpath(joinpath(raw_data_dir, dir))
                mkpath(dest_dir)
                dest_file = joinpath(dest_dir, basename(file))
                src_file = normpath(file)
                if dest_file != src_file
                    open(src_file, "r") do f
                        open(dest_file, "w") do g
                            write(g, read(f))
                        end
                    end
                end
            end
        end
    else
        file = abspath(inputdata)
        files = filter([file])
        for file in files
            @info "Pushing file to backend" backend=basename(tagger.backend_dir) instance=tagger.model_instance file=file
            dest_file = normpath(joinpath(raw_data_dir, basename(file)))
            isnothing(destination) && (destination = dest_file)
            src_file = normpath(file)
            if dest_file != src_file
                open(src_file, "r") do f
                    open(dest_file, "w") do g
                        write(g, read(f))
                    end
                end
            end
        end
    end
    return destination
end

function filter(files::Vector{String})
    files′= collect(files) # copy
    for file in files
        dir_stem, ext = splitext(file)
        if ext == ".spine"
            file′= dir_stem * ".outline"
            if isfile(file′)
                push!(files′, file′)
            end
        elseif ext == ".outline"
            file′= dir_stem * ".spine"
            if isfile(file′)
                push!(files′, file′)
            end
        else
            for file′ in Formats.getdependencies(file)
                push!(files′, file′)
            end
        end
    end
    return files′
end

function pull(tagger::Tagger, dest_dir::String)
    proc_data_dir = datadir(tagger, "processed")
    isdir(proc_data_dir) || throw("no processed data directory found")
    dest_dir = realpath(dest_dir) # strip end slash
    dest_files = String[]
    for (parent, _, files) in walkdir(proc_data_dir)
        if !isempty(files)
            parent′= joinpath(dest_dir, relpath(parent, proc_data_dir))
            for file in files
                src_file = normpath(joinpath(parent, file))
                dest_file = normpath(joinpath(parent′, file))
                if dest_file != src_file
                    open(src_file, "r") do f
                        open(dest_file, "w") do g
                            write(g, read(f))
                        end
                    end
                    check_permissions(dest_file)
                end
                push!(dest_files, dest_file)
            end
        end
    end
    return dest_files
end

"""
    check_permissions(filepath)

Attempt to fix file ownership.
This is useful in a Docker container that runs as root.
Environment variables HOST_UID and HOST_GID are set in the container by the larvatagger.sh script
so that the actual user is known.
"""
function check_permissions(file)
    try
        uid, gid = ENV["HOST_UID"], ENV["HOST_GID"]
        @debug "Changing file ownership" uid gid file
        chown(file, parse(Int, uid), parse(Int, gid))
    catch
    end
end

function parsekwargs!(args, kwargs)
    for (key, value) in pairs(kwargs)
        isnothing(value) && continue
        value === false && continue
        push!(args, "--" * replace(string(key), '_' => '-'))
        value === true && continue
        push!(args, value)
    end
    return args
end

function train(tagger::Tagger; pretrained_instance=nothing, kwargs...)
    args = ["--model-instance", tagger.model_instance]
    if !isnothing(pretrained_instance)
        push!(args, "--pretrained-model-instance")
        push!(args, pretrained_instance)
    end
    parsekwargs!(args, kwargs)
    ret = run(Cmd(`poetry run tagging-backend train $(args)`; dir=tagger.backend_dir))
    @assert isdir(modeldir(tagger))
    return ret
end

function predict(tagger::Tagger; kwargs...)
    args = ["--model-instance", tagger.model_instance]
    parsekwargs!(args, kwargs)
    run(Cmd(`poetry run tagging-backend predict $(args)`; dir=tagger.backend_dir))
end

end # module
