using DocOpt
using Pkg
using PlanarLarvae.Datasets, PlanarLarvae.Formats
using LarvaTagger, LarvaTagger.Taggers
using JSServe: JSServe, Server

usage = """Larva Tagger.

Usage:
  larvatagger.jl import <input-path> [<output-file>] [--id=<id>] [--framerate=<fps>]
  larvatagger.jl open <file-path> [--backends=<path>] [--port=<number>] [--quiet] [--viewer] [--browser]
  larvatagger.jl train <backend-path> <data-path> <model-instance> [--pretrained-model=<instance>] [--labels=<comma-separated-list>] [--sample-size=<N>] [--layers=<L>]
  larvatagger.jl predict <backend-path> <model-instance> <data-path> [--skip-make-dataset]
  larvatagger.jl -V | --version
  larvatagger.jl -h | --help

Options:
  -h --help            Show this screen.
  -V --version         Show version.
  -q --quiet           Do not show instructions.
  --id=<id>            Run ID, e.g. `date_time`.
  --framerate=<fps>    Camera frame rate, in frames per second.
  --backends=<path>    Path to backend repository.
  --port=<number>      Port number the server listens to.
  --viewer             Disable editing capabilities.
  --browser            Automatically open a browser tab at the served location.
  --skip-make-dataset  Skip the make_dataset stage; makes `predict` faster.
  --sample-size=<N>    Sample only N track segments from the data repository.
  --layers=<L>         (MaggotUBA) Number of layers of the classifier.
  --labels=<comma-separated-list>  Comma-separated list of behavior tags/labels.
  --pretrained-model=<instance>    Name of the pretrained encoder (from `pretrained_models` registry).

Note:
  <data-path> can be a path to a file or directory. In the case of `predict`,
  <data-path> can also be a .txt file listing data files; one relative path per line.

"""

function main(args=ARGS; exit_on_error=true)
    parsed_args = try
        docopt(usage, args isa String ? split(args) : args;
               help=false, exit_on_error=exit_on_error)
    catch # for exit_on_error==false only
        print("Parsing error -- ") # help message follows
        Dict{String, Any}()
    end
    if isempty(parsed_args) || parsed_args["--help"]
        print(usage)
    elseif parsed_args["--version"]
        println(Pkg.project().version)
    elseif parsed_args["import"]
        infile = parsed_args["<input-path>"]
        if !isfile(infile)
            @error "File not found; did you specify a file path after \"import\"?" infile
            return
        end
        outfile = parsed_args["<output-file>"]
        kwargs = Dict{Symbol, Any}()
        framerate = parsed_args["--framerate"]
        if !isnothing(framerate)
            kwargs[:framerate] = parse(Float64, framerate)
        end
        file = load(infile; kwargs...)
        runid = parsed_args["--id"]
        run = Run(isnothing(runid) ? file.run.id : runid, Track[];
                  file.run.attributes[:metadata]...)
        run.attributes[:labels] = String[]
        Datasets.pushdependency!(run, infile)
        if isnothing(outfile)
            Datasets.write_json(stdout, run)
        else
            outfile = joinpath(dirname(infile), outfile)
            Datasets.to_json_file(outfile, run)
            Taggers.check_permissions(outfile)
        end
    elseif parsed_args["open"]
        verbose = !parsed_args["--quiet"]
        infile = parsed_args["<file-path>"]
        if !isfile(infile)
            @error "File not found; did you specify a file path after \"open\"?" infile
            exit()
        end
        if parsed_args["--viewer"]
            app = larvaviewer(infile)
        else
            kwargs = Dict{Symbol, Any}()
            backends = parsed_args["--backends"]
            if !isnothing(backends)
                kwargs[:backend_directory] = backends
            end
            app = larvaeditor(infile; kwargs...)
        end
        #
        port = parsed_args["--port"]
        port = isnothing(port) ? 9284 : parse(Int, port)
        server = Server(app, "0.0.0.0", port)
        if parsed_args["--browser"]
            JSServe.openurl("http://127.0.0.1:$(port)")
        elseif verbose
            @info "The server is ready at http://127.0.0.1:$(port)"
        end
        if verbose
            @info "Press Ctrl+D to stop the server"
        end
        return server
    elseif parsed_args["train"]
        backend_path = parsed_args["<backend-path>"]
        model_instance = parsed_args["<model-instance>"]
        data_path = parsed_args["<data-path>"]
        #
        tagger = Tagger(backend_path, model_instance)
        Taggers.reset(tagger)
        Taggers.push(tagger, data_path)
        kwargs = Dict{Symbol, Any}()
        layers = parsed_args["--layers"]
        isnothing(layers) || (kwargs[:layers] = layers)
        train(tagger;
              pretrained_instance=parsed_args["--pretrained-model"],
              labels=parsed_args["--labels"],
              sample_size=parsed_args["--sample-size"],
              kwargs...
             )
    elseif parsed_args["predict"]
        backend_path = parsed_args["<backend-path>"]
        model_instance = parsed_args["<model-instance>"]
        data_path = parsed_args["<data-path>"]
        #
        tagger = Tagger(backend_path, model_instance)
        resetdata(tagger)
        datapath = abspath(data_path)
        Taggers.push(tagger, datapath)
        predict(tagger; skip_make_dataset=parsed_args["--skip-make-dataset"])
        Taggers.pull(tagger, isfile(datapath) ? dirname(datapath) : datapath)
    end
end
