
abstract type AbstractAnimator end

function stop! end
function play! end
function pause! end
function playpause! end
function isplaying end

"""
Plans are to make `Animator` not mutable; don't rely on its mutability.
"""
mutable struct Animator <: AbstractAnimator
    times::Vector{PlanarLarvae.Time}
    timestep::Observable
    speed::Observable
    isplaying::Observable
end

function playcontroller end

function bind! end

function playercontrols end

## implementation

const O{T} = Union{T, Observable{T}}
asobservable(arg) = Observable(arg)
asobservable(arg::Observable) = arg

Animator(times::Vector{PlanarLarvae.Time},
               step::O{Int}=1,
               speed::O{<:AbstractFloat}=1.0,
              ) = Animator(times,
                           asobservable(step),
                           asobservable(speed),
                           Observable(false),
                          )

isplaying(controller::AbstractAnimator) = controller.isplaying[]
isplaying!(controller::AbstractAnimator, state::Bool) = (controller.isplaying[] = state)

function playnow!(controller::AbstractAnimator)
    @info "Playing"
    ts = controller.times
    step = controller.timestep[]
    stop = length(ts)
    #
    lag, n = 0, 0
    clock_t = time()
    #
    while step<stop
        t, t′ = ts[step], ts[step + 1]
        dt = t′ - t
        dt′= dt / controller.speed[] - lag
        0 < dt′ && sleep(dt′)
        #
        step′ = ObservationPolicies.input(controller.timestep)[]
        if isplaying(controller)
            if step == step′
                step′ = step + 1
                @async controller.timestep[] = step′
            end
        else
            break
        end
        # adjust lag correction
        clock_t′ = time()
        lag′= clock_t′- clock_t - dt
        lag = (lag * n + lag′) / (n + 1)
        @debug "Step timing" expected=dt actual=clock_t′-clock_t adjustment=lag
        #
        step = step′
        clock_t = clock_t′
        n += 1
    end
    @info "Pause/stop"
    pause!(controller)
end

function bind!(controller::Animator)
    on(controller.isplaying) do playnow
        if playnow
            @async playnow!(controller)
        end
    end
    return controller
end

function bind!(controller::Animator, timeslider::Makie.Slider)
    controller.timestep = timeslider.selected_index
    #bind!(controller)
end

function bind!(controller::Animator, scene)
    on(events(scene).window_open) do isopen
        if !isopen
            stop!(controller)
        end
    end
    return controller
end

play!(controller::AbstractAnimator) = (controller.isplaying[] = true)
pause!(controller::AbstractAnimator) = (controller.isplaying[] = false)
playpause!(controller::AbstractAnimator) = (controller.isplaying[] = !controller.isplaying[])
stop!(controller) = pause!(controller)

function playcontroller(times; step::O{Int}=1, speed::O{<:AbstractFloat}=1.0)
    controller = Animator(times, step, speed)
    bind!(controller)
    return controller
end

gettime(controller::AbstractAnimator, step::Int) = controller.times[step]

struct TimeController <: AbstractAnimator
    times::Vector{Float64}
    time::AbstractObservable{Float64}
    timestep::AbstractObservable{Int64}
    stepmin::AbstractObservable{Int64}
    stepmax::AbstractObservable{Int64}
    boundreached::AbstractObservable{Bool}
    playing::AbstractObservable{Bool}
    speed::AbstractObservable{Float64}
end

isplaying(controller::TimeController) = controller.playing[]
play!(controller::TimeController) = (controller.playing[] = true)
pause!(controller::TimeController) = (controller.playing[] = false)
playpause!(controller::TimeController) = (controller.playing[] = !controller.playing[])

function stepforward!(controller::TimeController)
    step = controller.timestep[]
    if step < controller.stepmax[]
        controller.timestep[] = step + 1
    elseif isplaying(controller)
        pause!(controller)
    end
end

function stepbackward!(controller::TimeController)
    step = controller.timestep[]
    if controller.stepmin[] < step
        controller.timestep[] = step - 1
    elseif isplaying(controller)
        pause!(controller)
    end
end

function timecontroller(times::Vector{Float64}; speed=1.0)
    issorted(times) || throw(ArgumentError("timestamps are not sorted"))
    initialstep = 1
    timestep = newobservable(Cooldown(0.1), initialstep)
    time = Observable(times[initialstep])
    nsteps = length(times)
    stepmin = Observable(1)
    stepmax = Observable(nsteps)
    boundreached = Observable(true)
    #
    on(timestep) do step
        min, max = stepmin[], stepmax[]
        if step < min
            timestep.val = min
            @warn "step < stepmin"
        elseif max < step
            timestep.val = max
            @warn "step > stepmax"
        else
            time[] = times[step]
            if step == min || step == max
                boundreached[] = true
            elseif boundreached[]
                boundreached[] = false
            end
        end
    end
    on(boundreached) do b
        b && @info "Time bound reached"
    end
    on(stepmin) do bound
        0 < bound || throw(DomainError("stepmin < 1"))
        if timestep[] < bound
            timestep[] = bound
        end
    end
    on(stepmax) do bound
        bound <= nsteps || throw(DomainError("stepmax > #steps=$(nsteps)"))
        if bound < timestep[]
            timestep[] = bound
        end
    end
    #
    playing = Observable(false)
    if !(speed isa AbstractObservable)
        speed = Observable(speed)
    end
    controller = TimeController(times, time, timestep,
                                stepmin, stepmax, boundreached,
                                playing, speed)
    on(playing) do playnow
        if playnow
            @async playnow!(controller)
        end
    end
    return controller
end

function slave(master::TimeController, policy::ObservationPolicy=IndependentObservables)
    slave_timestep = newobservable(policy, master.timestep)
    slave_time = map(slave_timestep) do step
        master.times[step]
    end
    TimeController(master.times,
                   slave_time,
                   slave_timestep,
                   master.stepmin,
                   master.stepmax,
                   master.boundreached,
                   master.playing,
                   master.speed)
end

isslave(controller::TimeController) = ObservationPolicies.haspolicy(controller.timestep)

function stop!(controller::TimeController)
    pause!(controller)
    if isslave(controller)
        # duck typing
        try
            ObservationPolicies.stop!(controller.timestep.policy)
        catch
            @debug begin
                typ = typeof(controller.timestep.policy)
                "`stop!` not implemented for slave controller $(typ)"
            end
        end
    end
end

gettimes(c::AbstractAnimator) = c.times
gettimestep(c::AbstractAnimator) = c.timestep
settimestep!(c::AbstractAnimator, t) = (c.timestep[] = t)

gettimes(c) = gettimes(getplayer(c))
gettimestep(c) = gettimestep(getplayer(c))
settimestep!(c, t) = settimestep!(getplayer(c), t)

function settimebounds!(c::TimeController, larva)
    c.stepmin[] = larva.alignedsteps[1]
    c.stepmax[] = larva.alignedsteps[end]
    @debug "Bounding time" stepmin=c.stepmin[] stepmax=c.stepmax[]
end
function unsettimebounds!(c::TimeController)
    c.stepmin[] = 1
    c.stepmax[] = length(c.times)
    @debug "Unbounding time" stepmin=c.stepmin[] stepmax=c.stepmax[]
end
