getsession(c) = gethub(c)[:session]
setsession!(c, session) = (gethub(c)[:session] = session)

const r = JSServe.jsrender

const LarvaTaggerJS = JSServe.Dependency(:LarvaTagger,
                                         [joinpath(@__DIR__, "larvatagger.js")])
const LarvaTaggerCSS = JSServe.Asset(joinpath(@__DIR__, "larvatagger.css"))
const LoadAwesomeCSS = JSServe.Asset(joinpath(@__DIR__, "loadawesome.css"))

scrollable(elements...) = DOM.div(DOM.table(elements...);
                                  class="scrollable", scrolling="auto")

const Trigger = Observable{Bool}
const Toggle = Observable{Bool}

function simpletrigger(cls::DataType, args...)
    click = Trigger()
    on(click) do _
        #click.val = false
        @debug "Click on $cls"
    end
    return cls(click, args...)
end

function with_attributes(kwargs; class="", style="", kwargs′...)
    attributes = Dict{Symbol, Any}(kwargs′)
    merge!(attributes, Dict{Symbol, Any}(kwargs))
    if !isempty(class)
        existing = get(attributes, :class, "")
        attributes[:class] = isempty(existing) ? class : join((class, existing), " ")
    end
    if !isempty(style)
        existing = get(attributes, :style, "")
        if isempty(existing)
            style = endswith(style, ";") ? style : style * ";"
        else
            style = join((style, existing), endswith(style, ";") ? " " : "; ")
        end
        attributes[:style] = style
    end
    return attributes
end

css(lines...) = join(lines, "; ")

css_button = "button"

function DOMtrigger(label, trigger, attributes)
    DOM.button(label;
               type="button",
               onclick=js"""
               JSServe.update_obs($trigger, true);
               """,
               attributes...
              )
end

DOMtrigger(label, trigger; kwargs...) = DOMtrigger(label, trigger, with_attributes(kwargs; class=css_button))

function DOMtoggle(label, jlstate, attributes)
    DOM.button(label;
               type="button",
               onclick=js"LarvaTagger.toggle($jlstate)",
               attributes...
              )
end

struct ControlPanel
    id::String
    title::String
    active::Bool
    elements
    attributes
end

function ControlPanel(elements...; id::String, title::String, active::Bool=false, attributes...)
    attributes = with_attributes(attributes; class="control-panel")
    ControlPanel(id, title, active, elements, attributes)
end

function JSServe.jsrender(session::Session, panel::ControlPanel)
    tabid = "select-" * panel.id
    escaped_attrs = (:for=>tabid,)
    r(session,
      DOM.div(DOM.input(; type="radio", name="tab", id=tabid, class="cp-tab-switch",
                          checked=panel.active),
              DOM.label(panel.title; class="cp-tab-label", escaped_attrs...),
              DOM.div(panel.elements...; id=panel.id, panel.attributes...);
              class="cp-tab"))
end

cp(panels...) = DOM.div(panels...; class="cp-tabs")

struct HomeButton
    click::Trigger
    attributes
end

function homebutton(kwargs...)
    simpletrigger(HomeButton,
                  with_attributes(kwargs;
                                  class=css_button,
                                  style="position:absolute; top:1.5625rem; right:2.5rem;"))
end

function JSServe.jsrender(session::Session, b::HomeButton)
    r(session,
      DOMtrigger("⌂", b.click, b.attributes))
end

struct AssayPlot
    controller
    figure::Figure
    axis::Axis
    larvae::DecoratedLarvae
    homebutton::HomeButton
end

function AssayPlot(ctrl, larvae::DecoratedLarvae; size=FIGSIZE)
    fig = Figure(resolution=size)
    width = 0.1f0 # try to get 1px
    color = RGBAf(0, 0, 0, 0.36)
    ax = Axis(fig.layout[1, 1], aspect=DataAspect(), xgridwidth=width, ygridwidth=width,
              xgridcolor=color, ygridcolor=color)
    autosize!(ax, size)
    plot = larvaplot!(ax, larvae)
    setkeyboardevents!(plot.parent, getplayer(ctrl))
    setbounds!(ax, ctrl, larvae)
    button = homebutton()
    on(button.click) do _
        setbounds!(ax, ctrl, larvae)
    end
    AssayPlot(ctrl, fig, ax, larvae, button)
end

function AssayPlot(ctrl, larvae::Vector{LarvaModel}, visible::AbstractObservable{Bool};
        downsampling_threshold=1e5, kwargs...)
    npoints = sum([length(larva.path) for larva in larvae])
    if downsampling_threshold <= npoints
        @info "Trajectory down-sampling active" npoints
        sample = downsampler()
        larvae = sample.(larvae)
    end
    plot = AssayPlot(ctrl,
                     DecoratedLarvae(larvae,
                                     gettags(ctrl),
                                     gettimestep(ctrl),
                                     visible);
                     kwargs...)
    return plot
end

function AssayPlot(ctrl, larvae::Dict{LarvaID, LarvaModel}, args...; kwargs...)
    AssayPlot(ctrl, collect(values(larvae)), args...; kwargs...)
end

function assayplot(data, ctrl, args...; kwargs...)
    plot = AssayPlot(ctrl, data, args...; kwargs...)
    setmouseevents!(plot; consume=true)
    return plot
end

function JSServe.jsrender(session::Session, p::AssayPlot)
    r(session,
      DOM.div(r(session, p.homebutton),
              p.figure;
              class="relative"))
end

function setmouseevents!(plot::AssayPlot; kwargs...)
    setmouseevents!(plot.axis.scene, plot.larvae, plot.controller; kwargs...)
end

struct TrackPlot
    controller
    figure
    axis::Axis
    larvae::Vector{LarvaModel}
    view::SingleLarvaView
    homebutton::HomeButton
end

function TrackPlot(controller, larvae::Vector{LarvaModel};
        size=FIGSIZE, editabletags::Bool=true)
    view = SingleLarvaView(larvae, controller; editabletags=editabletags)
    fig = Figure(resolution=size)
    width = 0.1f0 # try to get 1px
    color = RGBAf(0, 0, 0, 0.36)
    ax = Axis(fig.layout[1, 1], aspect=DataAspect(), xgridwidth=width, ygridwidth=width,
              xgridcolor=color, ygridcolor=color)
    autosize!(ax, size)
    larvaplot!(ax, view)
    setbounds!(ax, controller, larvae)
    on(getactivelarva(controller)) do id
        if !isnothing(id)
            for larva in larvae
                if id == larva.id
                    setbounds!(ax, controller, larva)
                    settimebounds!(controller, larva)
                    break
                end
            end
        end
    end
    button = homebutton()
    on(button.click) do _
        deactivatelarva!(controller)
    end
    TrackPlot(controller, fig, ax, larvae, view, button)
end

TrackPlot(ctrl, model::Dict{<:Integer, LarvaModel}, args...; kwargs...) = TrackPlot(ctrl, collect(values(model)), args...; kwargs...)

trackplot(model, controller; kwargs...) = TrackPlot(controller, model; kwargs...)

function JSServe.jsrender(session::Session, p::TrackPlot)
    # TODO: read client window size
    r(session,
      DOM.div(r(session, p.homebutton),
              p.figure;
              class="relative"))
end

struct PlayPauseButton
    playing::AbstractObservable{Bool}
    attributes::Dict{Symbol, Any}
end

function playpausebutton(ctrl::TimeController; kwargs...)
    PlayPauseButton(ctrl.playing, with_attributes(kwargs; class=css_button))
end

function JSServe.jsrender(session::Session, b::PlayPauseButton)
    r(session,
      DOMtoggle("Play/Pause", b.playing, b.attributes))
end

struct StepForwardButton
    click::Trigger
    attributes::Dict{Symbol, Any}
end

function stepforwardbutton(ctrl::TimeController; kwargs...)
    button = simpletrigger(StepForwardButton, with_attributes(kwargs; class=css_button))
    on(button.click) do _
        stepforward!(ctrl)
    end
    return button
end

function JSServe.jsrender(session::Session, b::StepForwardButton)
    r(session,
      DOMtrigger(">", b.click, b.attributes))
end

struct StepBackwardButton
    click::Trigger
    attributes::Dict{Symbol, Any}
end

function stepbackwardbutton(ctrl::TimeController; kwargs...)
    button = simpletrigger(StepBackwardButton, with_attributes(kwargs; class=css_button))
    on(button.click) do _
        stepbackward!(ctrl)
    end
    return button
end

function JSServe.jsrender(session::Session, b::StepBackwardButton)
    r(session,
      DOMtrigger("<", b.click, b.attributes))
end

struct TimeSlider
    master::TimeController
    attributes::Dict{Symbol, Any}
end

function timeslider(master::TimeController; kwargs...)
    TimeSlider(master,
               with_attributes(kwargs;
                               id="timeslider",
                               style="width: 100%",
                               class="ml-1 mr-1"))
end
timeslider(times::Vector{Float64}; kwargs...) = timeslider(timecontroller(times); kwargs...)

function JSServe.jsrender(session::Session, ts::TimeSlider)
    master = ts.master
    input = ObservationPolicies.input(master.timestep)
    output = ObservationPolicies.output(master.timestep)
    moveslider = js"LarvaTagger.timeSlider(this, $input, $(master.stepmin), $(master.stepmax)).moveto(value)"
    slider = DOM.input(type="range",
                       min=1,
                       max=length(master.times),
                       value=output,
                       step=1,
                       oninput=moveslider;
                       ts.attributes...)
    JSServe.onload(session, slider, js"(slider)=>{ slider.value = 1; }")
    r(session, slider)
end

struct TimeLabel
    time::AbstractObservable{Float64}
    ndecimals::Int
    attributes::Dict{Symbol, Any}
end

function timelabel(time::AbstractObservable{Float64};
                   ndecimals=2, kwargs...)
    TimeLabel(time,
              ndecimals,
              with_attributes(kwargs,
                              style=css("width: 12.5rem",
                                        "display: flex",
                                        "flex-direction: row",
                                        "align-items: center")))
end
timelabel(master::TimeController; kwargs...) = timelabel(master.time; kwargs...)
timelabel(slave; kwargs...) = timelabel(slave.master; kwargs...)

function JSServe.jsrender(session::Session, tl::TimeLabel)
    formatter = generate_formatter("t=% .$(tl.ndecimals)f s")
    r(session,
      DOM.label(lift(formatter, tl.time);
                tl.attributes...))
end

struct Player
    controller::TimeController
    timeslider::TimeSlider
    timelabel::TimeLabel
    playpausebutton::PlayPauseButton
    stepbackwardbutton::StepBackwardButton
    stepforwardbutton::StepForwardButton
    attributes::Dict{Symbol, Any}
end


player(times::Vector; kwargs...) = player(timecontroller(times); kwargs...)

player(controller; kwargs...) = player(getplayer(controller); kwargs...)

function player(animator::TimeController; kwargs...)
    Player(animator,
           timeslider(animator),
           timelabel(animator),
           playpausebutton(animator),
           stepbackwardbutton(animator),
           stepforwardbutton(animator),
           with_attributes(kwargs;
                           class="flex flex-row player",
                           style="width: 100%; height: 2.5rem"),
          )
end

function JSServe.jsrender(session::Session, p::Player)
    r(session,
      DOM.div(r(session, p.playpausebutton),
              r(session, p.stepbackwardbutton),
              r(session, p.stepforwardbutton),
              r(session, p.timeslider),
              r(session, p.timelabel);
              onmouseenter=js"LarvaTagger.focusOnTimeSlider(this)",
              p.attributes...
             ))
end

mutable struct AssayViewer
    plot::AssayPlot
    player::Player
    visible::AbstractObservable{Bool}
    dom
end

function assayviewer(controller)
    model = getlarvae(controller)
    visible = Observable(true)
    plot = assayplot(model, controller, visible)
    play = player(controller)
    on(visible) do b
        if b
            # refresh view
            notify(gettimestep(controller))
        end
    end
    AssayViewer(plot, play, visible, nothing)
end

function JSServe.jsrender(session::Session, av::AssayViewer)
    dom = DOM.div(r(session, av.plot),
                  r(session, av.player);
                  class="flex flex-col",
                  onmouseenter=js"LarvaTagger.focusOnTimeSlider(this)",
                 )
    av.dom = dom
    return r(session, dom)
end

struct LarvaInfo
    controller
    id::LarvaID
    reviewed::AbstractObservable{Bool}
    edited::AbstractObservable{Bool}
    included::AbstractObservable{Bool}
end

function larvainfo(controller, id)
    reviewed = Observable(false)
    edited = Observable(false)
    on(edited) do b
        if b
            #@assert reviewed[]
            reviewed[] || @warn "Larva $id not reviewed"
        else
            @info "Discarding changes for larva $id"
            filterevents(controller).discard_larva_edits[] = id
        end
    end
    # on(reviewed) do b
    #     if !b && edited[]
    #         @info "[design] shall we discard the changes?"
    #         edited[] = false
    #     end
    # end
    included = Observable(false)
    LarvaInfo(controller,
              id,
              reviewed,
              edited,
              included)
end

JSServe.jsrender(session::Session, li::LarvaInfo) = r(session, prerender(li))

function prerender(li::LarvaInfo)
    label = "#$(li.id)"
    discard_larva_edits = js"LarvaTagger.discardLarvaEdits(this, $(li.edited), $label)"
    reviewed_checkbox = DOM.input(type="checkbox",
                                  checked=li.reviewed,
                                  onchange=js"""
                                  JSServe.update_obs($(li.reviewed), this.checked);
                                  """)
    edited_checkbox = DOM.input(type="checkbox",
                                checked=li.edited,
                                # onchange is not triggered from Julia
                                onchange=discard_larva_edits)
    include = js"LarvaTagger.includeLarva(this, $(li.included))"
    included_checkbox = DOM.input(type="checkbox",
                                  checked=li.included,
                                  class="included",
                                  onchange=include)
    DOM.tr(DOM.td(label),
           DOM.td(reviewed_checkbox, style="text-align: center;"),
           DOM.td(edited_checkbox, style="text-align: center;"),
           DOM.td(included_checkbox, style="text-align: center;"),
          )
end

struct LarvaFilter
    controller
    entries::Vector{LarvaInfo}
    includeall::AbstractObservable{Bool}
    saveas::AbstractObservable{Bool}
    attributes
end

function larvafilter(controller::ControllerHub; kwargs...)
    larvafilter(getlarvae(controller), controller; kwargs...)
end

function larvafilter(larvae, controller::ControllerHub; kwargs...)
    entries = [larvainfo(controller, id) for id in keys(larvae)]
    on(getactivelarva(controller)) do id
        if !isnothing(id)
            for entry in entries
                if id == entry.id
                    entry.reviewed[] = true
                    break
                end
            end
        end
    end
    includeall = Observable(false)
    on(includeall) do b
        for entry in entries
            if entry.included[] != b
                entry.included[] = b
            end
        end
    end
    saveas = Observable(false)
    gethub(controller)[:larvafilter] = LarvaFilter(controller, entries, includeall, saveas, kwargs)
end

getlarvafilter(controller) = gethub(controller)[:larvafilter]

function JSServe.jsrender(session::Session, lf::LarvaFilter)
    outputfile = getoutputfile(lf.controller)
    disabled = map(!, lf.saveas)
    button = r(session,
               DOM.button("Save as...";
                          type="button",
                          id="saveas",
                          onclick=js"LarvaTagger.setOutputFilename($outputfile)",
                          disabled=disabled,
                          style="width: 100%",
                          class=css_button))
    includeall = js"LarvaTagger.includeAllLarvae(this, $(lf.includeall), $button)"
    table = scrollable(DOM.tr(DOM.th(""),
                              DOM.th("Reviewed"),
                              DOM.th("Edited"),
                              DOM.th("Export")),
                       DOM.tr(DOM.td("All"),
                              DOM.td(),
                              DOM.td(),
                              DOM.td(DOM.input(type="checkbox",
                                               #checked=false,
                                               id="includeall",
                                               onchange=includeall),
                                     style="text-align: center;")),
                       prerender.(lf.entries)...)
    JSServe.onload(session, table, js"LarvaTagger.uncheckAllCheckboxes")
    r(session,
      DOM.div(table,
              button;
              with_attributes(lf.attributes; class="panel")...))
end

abstract type AbstractView end

mutable struct TagController
    hub::ControllerHub
    view
    fallback_color::String
end

function TagController(controller::ControllerHub,
        fallback_color::Union{Nothing, Symbol, String}=nothing,
    )
    taglut = gettags(controller)
    controller[:taghooks] = taghooks = newcontroller(taglut[], taglut)
    for tag in taghooks
        setevents!(controller, tag)
    end
    #
    fallback_color = @something fallback_color theme[:LarvaPlot][:fallback_color]
    TagController(controller,
                  nothing,
                  html_color(fallback_color))
end

function setcallbacks!(controller, tag, checkbox, label, colorpicker)
    session = getsession(controller)
    JSServe.attribute_render(session, label, "value", tag.name)
    JSServe.attribute_render(session, colorpicker, "value", tag.color)
    #
    on(tagevents(controller).renamefailed) do (name, failure)
        name == tag.name[] || return
        @info "Reverting to \"$(name)\" after failure: \"$(failure)\""
        evaljs(session,
               js"""
               const label = $(JSServe.selector(label));
               label.value = $(name);
               """)
    end
end

newtagname(::TagController) = NoTag

struct IndividualTagView
    controller::TagController
    tag::ObservableTag # view
end

tagview(controller, tag) = IndividualTagView(controller, tag)

function JSServe.jsrender(session::Session, ti::IndividualTagView)
    tag = ti.tag
    active = tag.active
    disabled = map(!, active)
    name = tag.name
    color = tag.color
    #
    label = DOM.input(type="text",
                      value=name[],
                      onchange=js"""
                      JSServe.update_obs($(name), this.value);
                      """,
                      disabled=disabled,
                      class="tag-name",
                     )
    colorpicker = DOM.input(type="color",
                            value=color[],
                            onchange=js"""
                            JSServe.update_obs($(color), this.value);
                            """,
                            disabled=disabled,
                            class="h-8 w-8 rounded shadow tag-color",
                           )
    checkbox = DOM.input(type="checkbox",
                         checked=active,
                         onchange=js"""
                         JSServe.update_obs($(active), this.checked);
                         """,
                         class="tag-active")
    setcallbacks!(ti.controller, tag, checkbox, label, colorpicker)
    notify(active)
    r(session,
      DOM.tr(DOM.td(checkbox),
             DOM.td(label),
             DOM.td(colorpicker),
            ))
end

function addtag! end

mutable struct AddTagButton
    click::Trigger
    attributes::Dict{Symbol, Any}
end

function addtagbutton(controller; kwargs...)
    button = simpletrigger(AddTagButton,
                           with_attributes(kwargs;
                                           class=css_button,
                                           style="width: 100%; height: 2rem; padding: 0px",
                                          ))
    on(button.click) do _
        addtag!(controller)
    end
    return button
end

function JSServe.jsrender(session::Session, b::AddTagButton)
    r(session, DOMtrigger(" + ", b.click, b.attributes))
end

mutable struct TagFilter
    controller::TagController
    views::Vector{IndividualTagView}
    addtagbutton::AddTagButton
    domtable::Union{Nothing, DOM.Hyperscript.Node{DOM.Hyperscript.HTMLSVG}}
    attributes
end

function TagFilter(controller::TagController; kwargs...)
    TagFilter(controller,
              IndividualTagView[],
              addtagbutton(controller),
              nothing,
              kwargs,
             )
end

function tagfilter(controller)
    gethub(controller)[:tag] = tc = TagController(controller)
    tc.view = TagFilter(tc)
    return tc.view
end

function JSServe.jsrender(session::Session, tf::TagFilter)
    setsession!(tf.controller, session)
    tags = newview!(gethub(tf.controller)[:taghooks])
    for tag in reverse(tags)
        push!(tf.views,
              tagview(tf.controller,
                      tag))
    end
    tf.domtable = r(session, scrollable(tf.views...))
    r(session,
      DOM.div(tf.domtable,
              tf.addtagbutton;
              with_attributes(tf.attributes; class="panel")...))
end

function addtag!(parent::TagController)
    taglut = gettags(parent)
    newtag = ObservableTag(newtagname(parent), parent.fallback_color)
    pushfirst!(taglut[], newtag)
    #
    taghooks = gethub(parent)[:taghooks]
    tag = newcontroller(newtag, taglut)
    pushfirst!(taghooks, tag)
    setevents!(parent, tag)
    #
    # model ordering: highest priority is first, lowest is last
    view = parent.view
    childview = tagview(parent,
                        newview!(tag))
    addtag!(view, childview)
    notify(taglut)
end

function addtag!(view::TagFilter, newtag::IndividualTagView)
    session = getsession(view.controller)
    js_dom = r(session, newtag)
    html_dom = "<table><tbody>" * replace(string(js_dom), "'" => "\\'") * "</tbody></table>"
    evaljs(session, js"LarvaTagger.insertNewTag($(view.domtable), $html_dom)")
    push!(view.views, newtag)
    return js_dom
end

struct TagSelector
    controller::TagController
    tagnames::Observable{Vector{String}}
    selected::Observable{Vector{Tuple{String, Observable{Bool}}}}
    selectable::Bool
    showtoggle::Bool
end

function TagSelector(controller::TagController,
        selectable::Bool=true,
        multiple::Union{Nothing, Bool}=nothing,
    )
    tag_lut = gettags(controller)
    getnames(tags) = [tag.name[] for tag in tags if tag.active[]]
    tagnames = Observable(getnames(tag_lut[]))
    on(tag_lut) do tags
        tagnames[] = getnames(tags)
    end
    registered_observables = Observable{Bool}[]
    selectedtags = Observable(Tuple{String, Observable{Bool}}[])
    multitag = filterevents(controller).allow_multiple_tags
    if !isnothing(multiple)
        multitag[] = multiple
    end
    ctrl = TagSelector(controller,
                       tagnames,
                       selectedtags,
                       selectable,
                       isnothing(multiple))
    refresh_view = Observable(false)
    on(tagnames) do names
        for i in length(registered_observables)+1:length(names)
            newobs = Observable(false)
            on(newobs) do b
                @debug "Tag $(i) $(b ? "" : "un")activated"
            end
            push!(registered_observables, newobs)
        end
        selectedtags[] = collect(zip(names, registered_observables[1:length(names)]))
        notify(refresh_view)
    end
    activelarva = getactivelarva(controller)
    onany(gettimestep(controller),
          refresh_view,
          tag_lut,
          multitag,
         ) do timestep, _, _, multitag
        id = activelarva[]
        isnothing(id) && return
        if isempty(selectedtags[]) && any(tag -> tag.active[], tag_lut[])
            # Workaround: generate view if not done yet.
            # Note that the view is generated anyway, but empty,
            # which is fine when no tags are defined.
            # Notifying `tag_lut` actually triggers a bug if
            # empty, e.g. on loading Chore files.
            # This also triggers another bug if all tags are unchecked.
            notify(tag_lut)
            return
        end
        if multitag
            refresh_selected_tags(controller, id, timestep, tag_lut, selectedtags, selectable)
        else
            refresh_selected_tag(controller, id, timestep, tag_lut, selectedtags, selectable)
        end
    end
    on(multitag) do b
        @debug (b ? "Multiple tags" : "Single tag") * " per timestep"
    end
    on(filterevents(controller).discard_larva_edits) do id
        activelarva[] == id || return
        #notify(refresh_view) # applies only to the tag selector
        notify(gettimestep(controller)) # full refresh
    end
    return ctrl
end

function refresh_selected_tags(controller, id, timestep, taglut, selectedtags, selectable)
    larva = getlarva(controller, id)
    tags = getusertags(larva, timestep; lut=taglut)
    tagnames = [tag.name[] for tag in tags]
    jscode = String[]
    for (name, selected) in Observables.to_value(selectedtags)
        isempty(name) && continue
        selected′ = name in tagnames
        if !selectable || selected[] != selected′
            selected[] = selected′
            push!(jscode, "document.getElementById('$(name)').selected = $(selected′);")
        end
    end
    isempty(jscode) && return
    evaljs(getsession(controller),
           JSServe.JSCode([JSServe.JSString(join(jscode, "\n"))]))
end

function refresh_selected_tag(controller, id, timestep, taglut, selectedtags, selectable)
    larva = getlarva(controller, id)
    tags = getusertags(larva, timestep; lut=taglut)
    refresh_selected_tag(controller, tags, selectedtags, selectable)
end

function refresh_selected_tag(controller, tags::UserTags, selectedtags, selectable)
    for tag in tags
        if tag.active[]
            return refresh_selected_tag(controller, tag, selectedtags, selectable)
        end
    end
    refresh_selected_tag(controller, nothing, selectedtags, selectable)
end

function refresh_selected_tag(controller,
        tag::Union{Nothing, ObservableTag},
        selectedtags,
        selectable,
    )
    jscode = String[]
    for (name, selected) in Observables.to_value(selectedtags)
        isempty(name) && continue
        selected′ = !isnothing(tag) && name == tag.name[]
        if !selectable || selected[] != selected′
            selected[] = selected′
            push!(jscode, "document.getElementById('$(name)').selected = $(selected′);")
        end
    end
    isempty(jscode) && return
    evaljs(getsession(controller),
           JSServe.JSCode([JSServe.JSString(join(jscode, "\n"))]))
end

function toggletag!(controller, tagname, selected, selectedtags, selectable)
    selected[] = !selected[]
    multitag = filterevents(controller).allow_multiple_tags[]
    multitag || selected[] || @debug "Deselecting a tag in a single tag setting"
    #
    larva_id = getactivelarva(controller)[]
    larva = getlarva(controller, larva_id)
    timestep = gettimestep(controller)[]
    taglut = gettags(controller)[]
    tags = getusertags!(larva, timestep; lut=taglut)
    tag = lookup(taglut, tagname)
    isnothing(tag) && throw(KeyError(tagname))
    #
    if selected[]
        if multitag || isempty(tags)
            transaction = AddTag(larva_id, timestep, tagname)
            settag!(tags, tag)
        else
            transaction = OverrideTags(larva_id, timestep, tagname, collect(tags))
            settag!(tags, tag; single=true)
            #
            refresh_selected_tag(controller, tag, selectedtags, selectable)
        end
    else
        @assert !isempty(tags)
        transaction = RemoveTag(larva_id, timestep, tagname)
        deletetag!(tags, tag)
    end
    push!(transactions(controller), transaction)
    notify(larva.usertags)
end

function to_dom(selectedtags, selectable)
    DOM.select(DOM.option(tagname;
                          id=tagname,
                          value=tagname,
                          selected=selected[],
                          onmousedown=selectable ? js"LarvaTagger.toggleTagAtPointer(event)" : nothing,
                         ) for (tagname, selected) in selectedtags;
               multiple=true,
               size=1,
              )
end

function JSServe.jsrender(session::Session, ts::TagSelector)
    setsession!(ts.controller, session)
    if ts.selectable
        controller = ts.controller
        selectedtags = ts.selected
        fromjs = Observable("")
        on(fromjs) do actuatedtag
            for (tag, selected) in selectedtags[]
                if tag == actuatedtag
                    toggletag!(controller, tag, selected, selectedtags, ts.selectable)
                    flag_active_larva_as_edited(controller)
                    break
                end
            end
        end
        evaljs(session, js"LarvaTagger.setTagSelector($fromjs)")
    end
    taglist = r(session, DOM.div(to_dom(ts.selected[], ts.selectable); id="tag-selector"))
    on(ts.selected) do selected
        js_dom = JSServe.jsrender(session, to_dom(selected, ts.selectable))
        html_dom = replace(string(js_dom), "'" => "\\'")
        evaljs(session,
               js"""
               const select = $(taglist);
               select.innerHTML = $html_dom;
               """)
    end
    elements = [taglist]
    if ts.showtoggle
        multitag = filterevents(ts.controller).allow_multiple_tags
        checkbox = DOM.div(DOM.input(type="checkbox",
                                     checked=multitag,
                                     onchange=js"""
                                     JSServe.update_obs($(multitag), this.checked);
                                     """,
                                     class="m-1"),
                           DOM.label("Multiple tags per time step");
                           class="flex flex-row")
        push!(elements, checkbox)
    end
    JSServe.jsrender(session,
                     DOM.div(elements...;
                             class="flex flex-col",
                             onmouseenter=js"LarvaTagger.focusOnTimeSlider()"))
end

function tagselector(controller;
        editabletags::Bool=true,
        multipletags::Union{Nothing, Bool}=nothing,
    )
    c = gethub(controller)
    TagSelector(haskey(c, :tag) ? c[:tag] : TagController(c), editabletags, multipletags)
end

mutable struct TrackViewer
    plot::TrackPlot
    player::Player
    tagselector::TagSelector
    dom
end

function trackviewer(controller;
        editabletags::Bool=true,
        multipletags::Union{Nothing, Bool}=nothing,
    )
    model = getlarvae(controller)
    plot = trackplot(model, controller; editabletags=editabletags)
    play = player(controller)
    sele = tagselector(controller; editabletags=editabletags, multipletags=multipletags)
    TrackViewer(plot, play, sele, nothing)
end

function JSServe.jsrender(session::Session, tv::TrackViewer)
    dom = DOM.div(r(session, tv.plot),
                  r(session, tv.player),
                  r(session, tv.tagselector);
                  class="flex flex-col",
                  style="display: none;",
                  id="trackviewer",
                  onmouseenter=js"LarvaTagger.focusOnTimeSlider(this)",
                 )
    tv.dom = dom
    r(session, dom)
end

struct MetadataEditor
    controller
    max_entries::Int
end

MetadataEditor(controller) = MetadataEditor(controller, 10)

metadataeditor(controller) = MetadataEditor(controller)

function JSServe.jsrender(session::Session, me::MetadataEditor)
    table = getmetadatatable(me.controller, me.max_entries)
    dom = DOM.div(DOM.div(DOM.input(value=fieldname,
                                    type="text",
                                    class="headerfield",
                                    onchange=js"""
                                    JSServe.update_obs($(fieldname), this.value);
                                    """),
                          DOM.input(value=fieldvalue,
                                    type="text",
                                    class="valuefield",
                                    onchange=js"""
                                    JSServe.update_obs($(fieldvalue), this.value);
                                    """);
                          class="flex flex-row")
                  for (fieldname, fieldvalue) in pairs(table);
                  id="metadata-panel",
                  class="flex flex-col panel")
    r(session, dom)
end

struct FileMenu
    controller::ControllerHub
    menuselection::Observable{String}
    attributes
end

function filemenu(controller; kwargs...)
    selection = Observable("")
    menu = FileMenu(gethub(controller),
                    selection,
                    with_attributes(kwargs; size=5))
    wd = getworkingdir(controller)
    on(selection) do file
        if '/' in file
            @logmsg SecurityAlert "'/' character found in file name" file client=identifyclient(controller)
        elseif !isempty(file)
            cwd(wd, file)
        end
    end
    return menu
end

function JSServe.jsrender(session::Session, fm::FileMenu)
    wdcontent = getwdcontent(fm.controller)
    inputfile = getinputfile(fm.controller)
    select = r(session,
               DOM.select(multiple=true;
                          with_attributes(fm.attributes; class="directory-content")...))
    onjs(session, wdcontent, js"(directory_entries) =>
         LarvaTagger.updateSelectOptions($select, directory_entries, $(fm.menuselection))")
    wd = map(getpath(fm.controller)) do path
        @assert !isempty(path)
        if isempty(path) || path == "."
            "."
        else
            parts = splitpath(path)
            length(parts) == 1 ? "./" * parts[1] : ".../" * parts[end]
        end
    end
    dom = DOM.div(DOM.input(value=wd,
                            type="text",
                            readonly=true),
                  select;
                  id="filemenu-panel",
                  class="flex flex-col panel")
    notify(wdcontent) # populate the select element
    r(session, dom)
end

function globalrefresh(::Type{Session}, controller)
    @info "Refreshing the view"
    evaljs(getsession(controller), js"LarvaTagger.reloadPage()")
    #turn_load_animation_off(controller)
end

function setedited!(c::LarvaFilter, id)
    for entry in c.entries
        if entry.id == id
            entry.edited[] = true
            break
        end
    end
end

function addtosavequeue!(c::LarvaFilter, id)
    for entry in c.entries
        if entry.id == id
            entry.included[] = true
            break
        end
    end
    c.saveas[] = true
end

struct BackendMenu
    backends::Backends
    send2backend::Observable
end

function backendmenu(controller, location)
    backends = getbackends(controller, location)
    send2backend = Observable(false)
    on(send2backend) do _
        predict(backends, controller[:input][])
    end
    BackendMenu(backends, send2backend)
end

function JSServe.jsrender(session::Session, bs::BackendMenu)
    models = bs.backends.model_instances
    select_dom = DOM.select(DOM.option(model; value=model) for model in models[];
                         onchange=js"JSServe.update_obs($(bs.backends.model_instance), this.value)",
                         class=css_button)
    onjs(session, models, js"(options) => LarvaTagger.updateSelectOptions($select_dom, options)")
    disabled = map(isempty, models)
    dom = DOM.div(DOM.label("Backend"),
                  DOM.select(DOM.option(backend; value=backend) for backend in bs.backends.backends;
                             onchange=js"JSServe.update_obs($(bs.backends.active_backend), this.value)",
                             class=css_button),
                  DOM.label("Model instance"),
                  select_dom,
                  DOMtrigger("Autotag", bs.send2backend; disabled=disabled);
                  class="flex flex-col panel")
    r(session, dom)
end

struct LoadAnimation
    visible::Observable{Bool}
end

LoadAnimation() = LoadAnimation(Observable(true))

function loadanimation(controller)
    la = LoadAnimation()
    gethub(controller)[:loadanimation] = la.visible
    return la
end

function JSServe.jsrender(session::Session, la::LoadAnimation)
    dom = DOM.div(DOM.div();
                  id="load-animation",
                  class="la-ball-clip-rotate")
    onjs(session, la.visible, js"(b) => $(dom).style.display = (b ? 'block' : 'none')")
    r(session, dom)
end
